/** @file
 * @brief Config class implementation. This class represents the actual configuration.
 *
 *
 *
 * @copyright Intra2net AG
 * @license GPLv2
*/

#include "config.h"

#include "service_dhs.h"
#include "service_ods.h"
#include "service_dyndns.h"
#include "service_dyns.h"
#include "service_easydns.h"
#include "service_tzo.h"
#include "service_zoneedit.h"
#include "service_gnudip.h"

#include <time.h>
#include <iostream>
#include <fstream>

#include <boost/foreach.hpp>
#include <boost/filesystem.hpp>
#include <boost/regex.hpp>
#include <boost/algorithm/string.hpp>



namespace po = boost::program_options;
namespace fs = boost::filesystem;
namespace ba = boost::algorithm;

typedef boost::shared_ptr<boost::program_options::options_description> Options_descriptionPtr;

using namespace std;

/**
 * Default Constructor. Available command line and config file options with their default values are defined here.
 */
Config::Config(Logger::Ptr _log, Serviceholder::Ptr _serviceholder)
    : Log(_log)
    , ServiceHolder(_serviceholder)
    , DaemonMode(false)
    , Syslog(false)
    , EnableIPv6(false)
    , Loglevel(0)
    , ConfigPath("/etc/bpdyndnsd")
    , ProxyPort(0)
    , ExternalWarningLog("")
    , ExternalWarningLevel(0)
{
    // Available service description config options
    po::options_description opt_desc_service("Service description options");
    opt_desc_service.add_options()
        ("protocol",po::value<string>(),"The service protocol.")
        ("server",po::value<string>(),"Servername needed for gnudip protocol.")
        ("host",po::value<string>(),"The hostname to update.")
        ("login",po::value<string>(),"Login name.")
        ("password",po::value<string>(),"Corresponding password.")
        ("update_interval",po::value<int>()->default_value(-1),"Update interval in minutes.")
        ("max_updates_within_interval",po::value<int>()->default_value(-1),"How many updates can be made in one interval.")
        ("dns_cache_ttl",po::value<int>()->default_value(-1),"How long a dns record is valid.")
    ;

    // Available command line only options
    po::options_description opt_desc_cmd_only("Command line only options");
    opt_desc_cmd_only.add_options()
        ("help,?","Show help.")
        ("version,v","Show version.")
        ("config,c",po::value<string>()->default_value("/etc/bpdyndnsd"),"Set the config path.")
    ;

     // Available generic options. Valid on cmd or in config file.
    po::options_description opt_desc_generic("Generic config options");
    opt_desc_generic.add_options()
        ("daemon_mode",po::value<bool>()->default_value(false),"Run as system daemon.")
        ("loglevel",po::value<int>()->default_value(0),"Loglevel.")
        ("syslog",po::value<bool>()->default_value(false),"Use syslog facility.")
        ("enable_ipv6",po::value<bool>()->default_value(false),"Try to use IPv6.")
        ("webcheck_url",po::value<string>()->default_value(""),"Use this URL to determine IP.")
        ("webcheck_url_alt",po::value<string>()->default_value(""),"Use this alternative URL to determine IP.")
        ("http_proxy",po::value<string>(),"Use this proxy for all http requests.")
        ("http_proxy_port",po::value<int>(),"Port of the proxy.")
        ("external_warning_log",po::value<string>()->default_value(""),"External programm to pass warning log messages to.")
        ("external_warning_level",po::value<int>()->default_value(0),"Warning messages of which loglevel should be passed to external programm.")
    ;

    // Define valid command line parameters
    Options_descriptionPtr _opt_desc_cmd(new po::options_description("Command line options"));
    OptDescCmd = _opt_desc_cmd;
    _opt_desc_cmd.reset();
    OptDescCmd->add(opt_desc_cmd_only);
    OptDescCmd->add(opt_desc_generic);
    OptDescCmd->add(opt_desc_service);

    // Define valid config file options
    Options_descriptionPtr _opt_desc_conf_main(new po::options_description("Config file options"));
    OptDescConfMain = _opt_desc_conf_main;
    _opt_desc_conf_main.reset();
    OptDescConfMain->add(opt_desc_generic);

    // Define valid service file options
    Options_descriptionPtr _opt_desc_conf_service(new po::options_description("Service file options"));
    OptDescConfService = _opt_desc_conf_service;
    _opt_desc_conf_service.reset();
    OptDescConfService->add(opt_desc_service);
}


/**
 * Default Destructor
 */
Config::~Config()
{
}


/**
 * Parses the command line arguments and does the needed actions.
 * @param argc Command line argument number given to main.
 * @param argv[] Pointer to command line argument array given to main.
 * @return 0 if all is fine, -1 if not.
 */
int Config::parse_cmd_line(int argc, char *argv[])
{
    try
    {
        po::store(po::parse_command_line(argc, argv, *this->OptDescCmd), VariablesMap);
        po::notify(VariablesMap);

        if ( VariablesMap.count("help") )
        {
            Log->print_usage(OptDescCmd);
            return -1;
        }
        else if ( VariablesMap.count("version") )
        {
            Log->print_version();
            return -1;
        }

        // Create a service object if all needed options are set on the command line
        if ( VariablesMap.count("protocol") && VariablesMap.count("host") && VariablesMap.count("login") && VariablesMap.count("password") )
        {
            // Get the cmd parameter values for protocol host login and password
            string protocol = VariablesMap["protocol"].as<string>();
            string host = VariablesMap["host"].as<string>();
            string login = VariablesMap["login"].as<string>();
            string password = VariablesMap["password"].as<string>();

            protocol = ba::to_lower_copy(protocol);

            string server;
            if ( VariablesMap.count("server") )
                server = VariablesMap["server"].as<string>();

            int update_interval = 0;
            if ( VariablesMap.count("update_interval") )
                update_interval = VariablesMap["update_interval"].as<int>();

            int max_updates_within_interval = 0;
            if ( VariablesMap.count("max_updates_within_interval") )
                max_updates_within_interval = VariablesMap["max_updates_within_interval"].as<int>();

            int dns_cache_ttl = 0;
            if ( VariablesMap.count("dns_cache_ttl") )
                dns_cache_ttl = VariablesMap["dns_cache_ttl"].as<int>();

            Service::Ptr service = create_service(protocol,server,host,login,password,update_interval,max_updates_within_interval,dns_cache_ttl);
            if ( service )
            {
                ServiceHolder->add_service(service);
                Log->print_service_object("New Service object from command line options:", service->get_protocol(), service->get_hostname(), service->get_login() ,service->get_password(), service->get_update_interval(), service->get_max_updates_within_interval(), service->get_dns_cache_ttl() , service->get_actual_ip(), service->get_last_updates());
            }
            else
                return -1;
        }
        else if ( VariablesMap.count("protocol") || VariablesMap.count("host") || VariablesMap.count("login") || VariablesMap.count("password") )
        {
            Log->print_missing_cmd_service_option();
            Log->print_usage(OptDescCmd);
            return -1;
        }

        if ( VariablesMap.count("config") )
        {
            fs::path full_config_path = fs::system_complete(fs::path(VariablesMap["config"].as<string>()));
            ConfigPath = full_config_path.string();
            if ( !fs::exists(full_config_path) ||  !fs::is_directory(full_config_path) )
            {
                // Config path doesn't exist or is not a directory
                Log->print_error_config_path(ConfigPath);
                return -1;
            }
        }

        if ( VariablesMap.count("daemon_mode") )
            DaemonMode = VariablesMap["daemon_mode"].as<bool>();

        if ( VariablesMap.count("loglevel") )
            Loglevel = VariablesMap["loglevel"].as<int>();

        if ( VariablesMap.count("syslog") )
            Syslog = VariablesMap["syslog"].as<bool>();

        if ( VariablesMap.count("enable_ipv6") )
            EnableIPv6 = VariablesMap["enable_ipv6"].as<bool>();

        if ( VariablesMap.count("webcheck_url") )
            WebcheckIpUrl = VariablesMap["webcheck_url"].as<string>();

        if ( VariablesMap.count("webcheck_url_alt") )
            WebcheckIpUrlAlt = VariablesMap["webcheck_url_alt"].as<string>();

        if ( VariablesMap.count("http_proxy") && VariablesMap.count("http_proxy_port") )
        {
            Proxy = VariablesMap["http_proxy"].as<string>();
            ProxyPort = VariablesMap["http_proxy_port"].as<int>();
        }
        else if ( VariablesMap.count("http_proxy") || VariablesMap.count("http_proxy_port") )
        {
            Log->print_missing_cmd_proxy_option();
            Log->print_usage(OptDescCmd);
            return -1;
        }

        if ( VariablesMap.count("external_warning_log") )
            ExternalWarningLog = VariablesMap["external_warning_log"].as<string>();

        if ( VariablesMap.count("external_warning_level") )
            ExternalWarningLevel = VariablesMap["external_warning_level"].as<int>();

    }
    catch(po::unknown_option e)
    {
        Log->print_unknown_cmd_option(e.what());
        Log->print_usage(OptDescCmd);
        return -1;
    }
    catch(po::multiple_occurrences e)
    {
        Log->print_multiple_cmd_option(e.what());
        Log->print_usage(OptDescCmd);
        return -1;
    }
    return 0;
}


/**
 * Creates a Service object from the given parameters.
 * @param protocol Protocol to use.
 * @param host Hostname to update.
 * @param login Login.
 * @param password Password.
 * @return A pointer to the created Service object.
 */
Service::Ptr Config::create_service(const string &protocol, const string& server, const string& hostname, const string& login, const string& password, const int update_interval, const int max_updates_within_interval, const int dns_cache_ttl)
{
    // Test for valid hostname. Must contain 3 parts minimum.
    list<string> fqhn_parts;
    ba::split(fqhn_parts,hostname,boost::is_any_of("."));
    if ( fqhn_parts.size() < 3 )
    {
        Log->print_invalid_hostname(protocol);
        Service::Ptr service;
        return service;
    }

    if(protocol == "dhs")
    {
        Service::Ptr service_dhs(new ServiceDhs(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_dhs;
    }
    else if(protocol == "ods")
    {
        Service::Ptr service_ods(new ServiceOds(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl));
        return service_ods;
    }
    else if(protocol == "dyndns")
    {
        Service::Ptr service_dyndns(new ServiceDyndns(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_dyndns;
    }
    else if(protocol == "dyns")
    {
        Service::Ptr service_dyns(new ServiceDyns(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_dyns;
    }
    else if(protocol == "easydns")
    {
        Service::Ptr service_easydns(new ServiceEasydns(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_easydns;
    }
    else if(protocol == "tzo")
    {
        Service::Ptr service_tzo(new ServiceTzo(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_tzo;
    }
    else if(protocol == "zoneedit")
    {
        Service::Ptr service_zoneedit(new ServiceZoneedit(protocol,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
        return service_zoneedit;
    }
    else if(protocol == "gnudip")
    {
        cout << "Server: " << server << endl; 
        if ( !server.empty() )
        {
            Service::Ptr service_gnudip(new ServiceGnudip(protocol,server,hostname,login,password,Log,update_interval,max_updates_within_interval,dns_cache_ttl,Proxy,ProxyPort));
            return service_gnudip;
        }
        else
        {
            Log->print_gnudip_requires_servername();
            Service::Ptr service;
            return service;
        }
    }
    else
    {
        Log->print_unknown_protocol(protocol);
        Service::Ptr service;
        return service;
    }
}


/**
 * Loads a service config file, invoked by load_config_from_files.
 * @param full_filename Filename of the service config file to load.
 * @return 0 if all is fine, -1 otherwise.
 */
int Config::load_service_config_file(const string& full_filename)
{
    Log->print_load_service_conf(full_filename);

    ifstream service_config_file(full_filename.c_str(),ifstream::in);
    if(service_config_file.is_open())
    {
        try
        {
            po::variables_map vm;
            po::parsed_options parsed_service_options = po::parse_config_file(service_config_file,*this->OptDescConfService,true);
            po::store(parsed_service_options,vm);
            po::notify(vm);

            if(vm.count("protocol") && vm.count("host") && vm.count("login") && vm.count("password"))
            {
                // create the corresponding service
                string protocol = vm["protocol"].as<string>();
                string host = vm["host"].as<string>();
                string login = vm["login"].as<string>();
                string password = vm["password"].as<string>();

                protocol = ba::to_lower_copy(protocol);

                string server;
                if ( vm.count("server") )
                    server = vm["server"].as<string>();

                int update_interval = 0;
                if ( vm.count("update_interval") )
                    update_interval = vm["update_interval"].as<int>();

                int max_updates_within_interval = 0;
                if ( vm.count("max_updates_within_interval") )
                    max_updates_within_interval = vm["max_updates_within_interval"].as<int>();

                int dns_cache_ttl = 0;
                if ( vm.count("dns_cache_ttl") )
                    dns_cache_ttl = vm["dns_cache_ttl"].as<int>();

                Service::Ptr service = create_service(protocol,server,host,login,password,update_interval,max_updates_within_interval,dns_cache_ttl);
                if ( service )
                {
                    ServiceHolder->add_service(service);
                    Log->print_service_object("New Service object from config:", service->get_protocol(), service->get_hostname(), service->get_login() ,service->get_password(), service->get_update_interval(), service->get_max_updates_within_interval(), service->get_dns_cache_ttl() , service->get_actual_ip(), service->get_last_updates());
                }
                else
                    return -1;
            }
            else if ( vm.count("protocol") || vm.count("host") || vm.count("login") || vm.count("password") )
            {
                service_config_file.close();
                Log->print_missing_service_conf_option(full_filename);
                return -1;
            }
        }
        catch ( po::unknown_option e )
        {
            // unknown option in config file detected
            service_config_file.close();
            Log->print_unknown_service_conf_option(full_filename,e.what());
            return -1;
        }
        catch(po::multiple_occurrences e)
        {
            service_config_file.close();
            Log->print_multiple_service_conf_option(full_filename,e.what());
            return -1;
        }
        service_config_file.close();
    }
    else
    {
        // error opening service config file for reading
        Log->print_error_opening_r(full_filename);
        return -1;
    }
    return 0;
}


/**
 * Loads the main config file, invoked by load_config_from_files
 * @param full_filename The full filename of the main config file to load
 * @return 0 if all is fine, -1 otherwise
 */
int Config::load_main_config_file(const string& full_filename)
{
    Log->print_load_main_conf(full_filename);

    ifstream main_config_file(full_filename.c_str(),ifstream::in);
    if(main_config_file.is_open())
    {
        try
        {
            po::parsed_options parsed_main_options = po::parse_config_file(main_config_file,*this->OptDescConfMain,true);
            po::store(parsed_main_options,VariablesMap);
            po::notify(VariablesMap);

            if ( VariablesMap.count("daemon_mode") )
                DaemonMode = VariablesMap["daemon_mode"].as<bool>();

            if ( VariablesMap.count("loglevel") )
                Loglevel = VariablesMap["loglevel"].as<int>();

            if ( VariablesMap.count("syslog") )
                Syslog = VariablesMap["syslog"].as<bool>();

            if ( VariablesMap.count("enable_ipv6") )
                EnableIPv6 = VariablesMap["enable_ipv6"].as<bool>();

            if ( VariablesMap.count("webcheck_url") )
                WebcheckIpUrl = VariablesMap["webcheck_url"].as<string>();

            if ( VariablesMap.count("webcheck_url_alt") )
                WebcheckIpUrlAlt = VariablesMap["webcheck_url_alt"].as<string>();

            if ( VariablesMap.count("http_proxy") && VariablesMap.count("http_proxy_port") )
            {
                Proxy = VariablesMap["http_proxy"].as<string>();
                ProxyPort = VariablesMap["http_proxy_port"].as<int>();
            }
            else if ( VariablesMap.count("http_proxy") || VariablesMap.count("http_proxy_port") )
            {
                main_config_file.close();
                Log->print_missing_conf_proxy_option(full_filename);
                return -1;
            }

            if ( VariablesMap.count("external_warning_log") )
                ExternalWarningLog = VariablesMap["external_warning_log"].as<string>();

            if ( VariablesMap.count("external_warning_level") )
                ExternalWarningLevel = VariablesMap["external_warning_level"].as<int>();

        }
        catch ( po::unknown_option e )      // at the moment 04-08-2009 this exception is never thrown :-(
        {
            // unknown option in main config file detected
            main_config_file.close();
            Log->print_unknown_main_conf_option(e.what());
            return -1;
        }
        catch(po::multiple_occurrences e)
        {
            main_config_file.close();
            Log->print_multiple_main_conf_option(full_filename,e.what());
            return -1;
        }
        main_config_file.close();
    }
    else
    {
        // error opening main config file for reading
        Log->print_error_opening_r(full_filename);
        return -1;
    }
    return 0;
}


/**
 * Loads the main and the service config file and does the needed action.
 * @param config_path The path to the config directory.
 * @return 0 if all is fine, -1 otherwise
 */
int Config::load_config_from_files()
{
    fs::path full_config_path = fs::path(ConfigPath);

    fs::directory_iterator end_iter;
    for ( fs::directory_iterator dir_itr(full_config_path) ; dir_itr != end_iter ; ++dir_itr )
    {
        if( fs::is_regular_file( dir_itr->status() ) )
        {
            string actual_file = dir_itr->path().filename();
            boost::regex expr(".*\\.conf?");
             // If it is the main config file do the following
            if ( actual_file == "bpdyndnsd.conf" )
            {
                // Load the main config file
                string full_filename = dir_itr->path().string();
                if ( load_main_config_file(full_filename) != 0 )
                    return -1;
            }
            // If it is a service definition file *.conf, parse it and generate the corresponding service
            else if ( boost::regex_search( actual_file,expr ) )
            {
                string full_filename = dir_itr->path().string();
                if ( load_service_config_file(full_filename) != 0 )
                    return -1;
            }
        }
    }
    // Config file successfully loaded
    Log->print_conf_loaded(ConfigPath);
    return 0;
}


/**
 * Getter method for member OptDescCmd.
 * @return options_description*.
 */
Options_descriptionPtr Config::get_opt_desc_cmd() const
{
    return OptDescCmd;
}


/**
 * Getter method for member OptDescConfMain.
 * @return options_description*.
 */
Options_descriptionPtr Config::get_opt_desc_conf_main() const
{
    return OptDescConfMain;
}


/**
 * Getter method for member OptDescConfService.
 * @return options_description*.
 */
Options_descriptionPtr Config::get_opt_desc_conf_service() const
{
    return OptDescConfService;
}


/**
 * Getter for member Loglevel.
 * @return Member Loglevel.
 */
int Config::get_loglevel() const
{
    return Loglevel;
}


/**
 * Getter for member DaemonMode.
 * @return TRUE if enabled, FALSE if disabled.
 */
bool Config::get_daemon_mode() const
{
    return DaemonMode;
}


/**
 * Deletes the map with the previously parsed options.
 * This is needed in case we reload the config and don't want the old cmd options to overwrite new config file options.
 */
void Config::delete_variables_map()
{
    VariablesMap.clear();

    po::variables_map _variables_map;
    VariablesMap = _variables_map;
}


/**
 * Getter for member Syslog.
 * @return True if logging through syslog is enabled, false otherwise.
 */
bool Config::get_syslog() const
{
    return Syslog;
}


/**
 * Getter for member EnableIPv6
 * @return Wether IPv6 should be used or not.
 */
bool Config::get_enable_ipv6() const
{
    return EnableIPv6;
}


/**
 * Getter for member WebcheckIpUrl
 * @return The primary IP Webcheck URL
 */
string Config::get_webcheck_ip_url() const
{
    return WebcheckIpUrl;
}


/**
 * Getter for member WebcheckIpUrlAlt
 * @return The alternative IP Webcheck URL
 */
string Config::get_webcheck_ip_url_alt() const
{
    return WebcheckIpUrlAlt;
}


/**
 * Get member Proxy
 * @return Proxy
 */
string Config::get_proxy() const
{
    return Proxy;
}


/**
 * Get member ProxyPort
 * @return ProxyPort
 */
int Config::get_proxy_port() const
{
    return ProxyPort;
}


/**
 * Get member ExternalWarningLog
 * @return ExternalWarningLog
 */
string Config::get_external_warning_log() const
{
    return ExternalWarningLog;
}


/**
 * Get member ExternalWarningLevel
 * @return ExternalWaringLevel
 */
int Config::get_external_warning_level() const
{
    return ExternalWarningLevel;
}

